"""The paint factory creates instances of QBrush, QPen and QFont."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Qt
from PySide6.QtGui import QColor, QPen
from vistutils.text import stringList
from vistutils.parse import maybe

from vistside.core import parseColor


def _createPen(*args, **kwargs) -> QPen:
  """Creates a QPen instance."""
  errors = []
  width, color, style = None, None, None
  widthDefault = 1
  colorDefault = QColor(0, 0, 0, 255)
  styleDefault = Qt.PenStyle.SolidLine
  capStyleDefault = Qt.PenCapStyle.FlatCap
  widthKeys = stringList("""width, w, penWidth""")
  colorKeys = stringList("""color, col, penColor, rgb, RGB, rgba, RGBA""")
  styleKeys = stringList("""style, penStyle, lineStyle""")
  capStyleKeys = stringList("""capStyle, penCapStyle""")
  widthArg, widthFArg, colorArg, styleArg = None, None, None, None
  widthKwarg, widthFKwarg, colorKwarg, styleKwarg = None, None, None, None
  capStyleArg, capStyleKwarg = None, None
  for arg in args:
    if isinstance(arg, float) and widthFArg is None:
      widthFArg = arg
    elif isinstance(arg, int) and widthArg is None:
      widthArg = arg
    elif isinstance(arg, QColor) and colorArg is None:
      colorArg = arg
    elif isinstance(arg, Qt.PenStyle) and styleArg is None:
      styleArg = arg
    elif isinstance(arg, Qt.PenCapStyle) and capStyleArg is None:
      capStyleArg = arg
  for key in widthKeys:
    if key in kwargs:
      if widthKwarg is None:
        val = kwargs[key]
        if isinstance(val, float):
          widthFKwarg = val
        if isinstance(val, int):
          widthKwarg = val
  for key in colorKeys:
    if key in kwargs:
      if colorKwarg is None:
        val = kwargs[key]
        try:
          colorKwarg = parseColor(val)
        except ValueError as valueError:
          errors.append(valueError)
  for key in styleKeys:
    if key in kwargs:
      if styleKwarg is None:
        val = kwargs[key]
        if isinstance(val, Qt.PenStyle):
          styleKwarg = val
  for key in capStyleKeys:
    if key in kwargs:
      if capStyleArg is None:
        val = kwargs[key]
        if isinstance(val, Qt.PenCapStyle):
          capStyleKwarg = val
  width = maybe(widthKwarg, widthArg, widthDefault)
  widthF = maybe(widthKwarg, widthArg, None)
  color = maybe(colorKwarg, colorArg, colorDefault)
  style = maybe(styleKwarg, styleArg, styleDefault)
  capStyle = maybe(capStyleKwarg, capStyleArg, capStyleDefault)
  pen = QPen()
  if widthF is not None:
    if isinstance(widthF, float):
      pen.setWidthF(widthF)
    elif isinstance(widthF, int):
      pen.setWidth(widthF)
  elif width is not None:
    pen.setWidth(width)
  pen.setColor(color)
  pen.setStyle(style)
  pen.setCapStyle(capStyle)
  return pen


def parsePen(*args, **kwargs) -> QPen:
  """Creates a QPen instance."""
  return _createPen(*args, **kwargs)


def emptyPen() -> QPen:
  """Creates a QPen instance."""
  return _createPen(QColor(0, 0, 0, 0, ), 1, Qt.PenStyle.NoPen)


def textPen(*args) -> QPen:
  """Creates a QPen instance."""
  color = None
  intArgs = []
  for arg in args:
    if isinstance(arg, QColor):
      color = arg
    if isinstance(arg, int):
      intArgs.append(arg)
    if isinstance(arg, str):
      color = parseColor(arg)
    if color is not None:
      break
  else:
    if len(intArgs) > 2:
      r, g, b, a = [*intArgs, 255][:4]
      color = QColor(r, g, b, a)
    else:
      color = QColor(0, 0, 0, 255)
  return _createPen(QColor(0, 0, 0, 255, ), 1, Qt.PenStyle.SolidLine)


def stylePen(penStyle: Qt.PenStyle, *args) -> QPen:
  """Creates a QPen instance."""
  color = maybe(parseColor(*args, strict=False), QColor(0, 0, 0, 255, ))
  return _createPen(color, 1, penStyle)


def solidPen(*args) -> QPen:
  """Creates a QPen instance."""
  return stylePen(Qt.PenStyle.SolidLine, *args)


def dashPen(*args) -> QPen:
  """Creates a QPen instance."""
  return stylePen(Qt.PenStyle.DashLine, *args)


def dotPen(*args) -> QPen:
  """Creates a QPen instance."""
  return stylePen(Qt.PenStyle.DotLine, *args)


def dashDotPen(*args) -> QPen:
  """Creates a QPen instance."""
  return stylePen(Qt.PenStyle.DashDotLine, *args)



"""TimerField provides QTimer typed descriptor class"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from PySide6.QtCore import QTimer
from icecream import ic
from vistutils.fields import Wait, unParseArgs
from vistutils.parse import maybe
from vistutils.text import stringList
from vistutils.waitaminute import typeMsg

from vistside.core import parseTimer, TimerType, Precise

ic.configureOutput(includeContext=True)


class Timer(QTimer):
  """Timer wraps the QTimer class allowing it to interact with the
  descriptor classes."""

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the descriptor"""
    QTimer.__init__(self, *args, **kwargs)

  @classmethod
  def getDefault(cls, *args, **kwargs) -> Any:
    """Returns the default value for the field."""
    defVal = cls()
    defVal.apply((args, kwargs))
    return defVal

  def apply(self, value: Any) -> Timer:
    """Applies the arguments contained in value to the timer."""
    args, kwargs = unParseArgs(value)
    epoch = None
    epochKeys = stringList("""epoch, time, interval, length, duration, d""")
    timerType = None
    timerKeys = stringList("""timer, type, precision, accuracy, acc""")
    singleShot = None
    singleShotKeys = stringList("""single, shot, one, once, only""")
    names = stringList('epoch, timerType, singleShot')
    types = [int, TimerType, bool]
    Keys = [epochKeys, timerKeys, singleShotKeys]
    parsedKwargs = {key: None for key in names}
    for (name, keys, type_) in zip(names, Keys, types):
      for key in keys:
        if key in kwargs and parsedKwargs[key] is None:
          val = kwargs.get(key)
          if isinstance(val, type_):
            parsedKwargs[name] = val
            break
          e = typeMsg(key, val, type_)
          raise TypeError(e)
    parsedArgs = {key: None for key in names}
    for (name, type_) in zip(names, types):
      for arg in args:
        if isinstance(arg, type_) and parsedArgs[name] is None:
          parsedArgs[name] = arg
          break
    parsedDefaults = dict(epoch=1000, timerType=Precise, singleShot=False)
    values = {}
    P = [parsedKwargs, parsedArgs, parsedDefaults]
    epoch = maybe(*[p.get('epoch') for p in P])
    if not isinstance(epoch, int):
      e = typeMsg('epoch', epoch, int)
      raise TypeError(e)
    timerType = maybe(*[p.get('timerType') for p in P])
    if not isinstance(timerType, TimerType):
      e = typeMsg('timerType', timerType, TimerType)
      raise TypeError(e)
    singleShot = maybe(*[p.get('singleShot') for p in P])
    if not isinstance(singleShot, bool):
      e = typeMsg('singleShot', singleShot, bool)
      raise TypeError(e)
    self.setInterval(epoch)
    self.setTimerType(timerType)
    self.setSingleShot(singleShot)
    return self


class TimerField(Wait):
  """TimerField provides QTimer typed descriptor class"""

  def __init__(self, *args, **kwargs) -> None:
    Wait.__init__(self, Timer, *args, **kwargs)



"""BaseWindow provides the menus and actions used by the main application
window. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import numpy as np
from PySide6.QtCore import Signal
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMainWindow, QMenuBar, QApplication, QStatusBar
from vistutils.fields import Wait

from vistside.core import generateWhiteNoise, TimerField, Precise
from vistside.widgets import PlotWidget, BaseLayoutField, BaseWidget, \
  DynamicPlotWidget
from vistside.windows.actions import FilesMenu, EditMenu, HelpMenu, \
  MainStatusBar
from vistside.windows.actions import DebugMenu


class BaseWindow(QMainWindow):
  """BaseWindow provides the menus and actions used by the main
  application window. """

  newAction: QAction
  openAction: QAction
  saveAction: QAction
  saveAsAction: QAction
  exitAction: QAction

  selectAllAction: QAction
  cutAction: QAction
  copyAction: QAction
  pasteAction: QAction
  undoAction: QAction
  redoAction: QAction

  aboutAction: QAction
  aboutQtAction: QAction
  aboutPythonAction: QAction

  debug01Action: QAction
  debug02Action: QAction
  debug03Action: QAction
  debug04Action: QAction
  debug05Action: QAction
  debug06Action: QAction
  debug07Action: QAction
  debug08Action: QAction
  debug09Action: QAction
  debug10Action: QAction

  __main_menu_bar__: QMenuBar = None
  __files_menu__: FilesMenu = None
  __edit_menu__: EditMenu = None
  __help_menu__: HelpMenu = None
  __debug_menu__: DebugMenu = None

  showSignal = Signal()

  mainStatusBar = Wait(MainStatusBar, )
  baseWidget = Wait(BaseWidget, )
  baseLayout = BaseLayoutField('vertical')
  paintTimer = TimerField(20, Precise, singleShot=False)

  def __init__(self, *args, **kwargs) -> None:
    QMainWindow.__init__(self)
    self.setMinimumSize(640, 480)
    self.setWindowTitle('Welcome to EZRos!')
    self.timePlot = None
    self.freqPlot = None
    self.__main_menu_bar__ = QMenuBar(self)
    self.__files_menu__ = FilesMenu(self, 'Files')
    self.__edit_menu__ = EditMenu(self, 'Edit')
    self.__help_menu__ = HelpMenu(self, 'Help')
    self.__debug_menu__ = DebugMenu(self, 'DEBUG')

  def show(self) -> None:
    """show displays the window."""
    self.setupMenus()
    self.connectActions()
    self.setStatusBar(self.mainStatusBar)
    self.baseWidget.setLayout(self.baseLayout)
    self.setCentralWidget(self.baseWidget)
    QMainWindow.show(self)
    self.showSignal.emit()

  def setupMenus(self, ) -> None:
    """Function responsible for setting up menus."""
    self.__main_menu_bar__.addMenu(self.__files_menu__)
    self.__main_menu_bar__.addMenu(self.__edit_menu__)
    self.__main_menu_bar__.addMenu(self.__help_menu__)
    self.__main_menu_bar__.addMenu(self.__debug_menu__)
    self.setMenuBar(self.__main_menu_bar__)
    self.__files_menu__.setupActions()
    self.__edit_menu__.setupActions()
    self.__help_menu__.setupActions()
    self.__debug_menu__.setupActions()

  def connectActions(self, ) -> None:
    """connectActions connects the actions to the status bar."""
    self.connectDebugActions()
    self.aboutQtAction.triggered.connect(QApplication.aboutQt)
    self.exitAction.triggered.connect(QApplication.quit)

  def connectDebugActions(self, ) -> None:
    """connectDebugActions connects the debug actions to the status bar."""
    self.debug01Action.triggered.connect(self.debug01Func)
    self.debug02Action.triggered.connect(self.debug02Func)
    self.debug03Action.triggered.connect(self.debug03Func)
    self.debug04Action.triggered.connect(self.debug04Func)
    self.debug05Action.triggered.connect(self.debug05Func)
    self.debug06Action.triggered.connect(self.debug06Func)
    self.debug07Action.triggered.connect(self.debug07Func)
    self.debug08Action.triggered.connect(self.debug08Func)
    self.debug09Action.triggered.connect(self.debug09Func)
    self.debug10Action.triggered.connect(self.debug10Func)

  def debug01Func(self, ) -> None:
    """debug01Func provides a debug function for the window."""
    print('Debug01 action triggered!')
    self.mainStatusBar.showMessage('Debug01 action triggered!')

  def debug02Func(self, ) -> None:
    """debug02Func provides a debug function for the window."""
    print('Debug02 action triggered!')

  def debug03Func(self, ) -> None:  # noqa
    """debug03Func provides a debug function for the window."""
    print('Debug03 action triggered!')

  def debug04Func(self, ) -> None:
    """debug04Func provides a debug function for the window."""
    print('Debug04 action triggered!')

  def debug05Func(self, ) -> None:
    """debug05Func provides a debug function for the window."""
    print('Debug05 action triggered!')

  def debug06Func(self, ) -> None:
    """debug06Func provides a debug function for the window."""
    print('Debug06 action triggered!')

  def debug07Func(self, ) -> None:
    """debug07Func provides a debug function for the window."""
    print('Debug07 action triggered!')

  def debug08Func(self, ) -> None:
    """debug08Func provides a debug function for the window."""
    print('Debug08 action triggered!')

  def debug09Func(self, ) -> None:
    """debug09Func provides a debug function for the window."""
    print('Debug09 action triggered!')

  def debug10Func(self, ) -> None:
    """debug10Func provides a debug function for the window."""
    print('Debug10 action triggered!')



"""The 'windows' module provides the window classes for the main
application window as well as dialogs."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._base_window import BaseWindow
from ._layout_window import LayoutWindow



"""EmptyPen provides a descriptor returning the empty pen"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Qt
from PySide6.QtGui import QPen, QColor
from vistutils.fields import StaticField


class EmptyPen(StaticField):
  """EmptyPen provides a descriptor returning the empty pen"""

  def __init__(self, ) -> None:
    pen = QPen()
    pen.setColor(QColor(0, 0, 0, 0, ))
    pen.setStyle(Qt.PenStyle.NoPen)
    StaticField.__init__(self, pen)



"""PlotWidget draws a static plot"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Self, Any

import numpy as np
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import (QVBoxLayout)
from PySide6.QtCharts import QChart, QChartView, QLineSeries, QValueAxis
from vistutils.fields import unParseArgs

from vistside.core import generateWhiteNoise, interpolateWhiteNoise
from vistside.widgets import BaseWidget


class PlotWidget(BaseWidget):
  """Displays a static plot using QChart."""

  def __init__(self, x_data: list, y_data: list):
    """
    Initializes the widget.

    Args:
        x_data: List of x-values for the plot.
        y_data: List of y-values for the plot.
    """
    super().__init__()

    self.chart = QChart()
    self.series = QLineSeries()
    self.chart.addSeries(self.series)

    self.set_data(x_data, y_data)

    # Customize axes (optional)
    axisX = QValueAxis()
    axisY = QValueAxis()
    self.chart.setAxisX(axisX, self.series)
    self.chart.setAxisY(axisY, self.series)

    # Customize appearance (optional)
    self.chart.legend().hide()

    self.chart_view = QChartView(self.chart)
    self.chart_view.setRenderHint(QPainter.Antialiasing)

    layout = QVBoxLayout()
    layout.addWidget(self.chart_view)
    self.setLayout(layout)

  def set_data(self, x_data: list, y_data: list) -> None:
    """Sets the data to be plotted."""
    self.series.clear()  # Clear existing data
    for x, y in zip(x_data, y_data):
      self.series.append(x, y)

  @classmethod
  def getDefault(cls, *args, **kwargs) -> Self:
    """Creates a default instance"""
    t = np.linspace(0, 2 * np.pi, 100)
    x = np.sin(t)
    return cls(t.tolist(), x.tolist())

  def apply(self, val: Any) -> Self:
    """Applies the value to the field."""
    args, kwargs = unParseArgs(val)
    x, y = None, None
    for arg in args:
      if isinstance(arg, np.ndarray):
        if x is None:
          x = arg
        elif y is None:
          y = arg
          break
    if x is None or y is None:
      return self
    self.series.appendNP(x, y)
    return self



"""AbstractPaint provides the abstract baseclass for the encapsulated
painting operations"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Self

from PySide6.QtGui import QPainter, QPaintEvent
from vistutils.fields import Field

if TYPE_CHECKING:
  from ezros.gui.widgets import PaintWidget

  PW = PaintWidget

Sig = tuple[type]


class AbstractPaint(Field):
  """AbstractPaint provides the abstract baseclass for the encapsulated
  painting operations"""

  @abstractmethod
  def paintOp(self, event: QPaintEvent, painter: QPainter) -> None:
    """Applies the paint operation"""

  def __set_name__(self, owner: PaintWidget, name: str) -> None:
    owner.appendHookedPaint(self)

  def __get__(self, instance: PW, owner: type[PW]) -> Self:
    return self



"""DataPaint paints data points on the widget"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from math import ceil, floor
from typing import Any

from PySide6.QtCore import Qt, QRect, QPoint, QLine
from PySide6.QtGui import QPainter, QPaintEvent

from ezros.gui.factories import emptyBrush, parsePen, parseFont, parseBrush
from ezros.gui.factories import emptyPen
from ezros.gui.paint import AbstractPaint
from ezros.gui.shortnames import Silver, Black, RoundCap, SolidLine
from ezros.gui.shortnames import AliceBlue, SolidFill, OliveDrab
from ezros.rosutils import Wait, DataEcho


class DataPaint(AbstractPaint):
  """DataPaint paints data points on the widget"""

  __fallback_pen__ = None

  pointPen = Wait(parsePen, Black, 4, SolidLine, RoundCap)
  majorGridPen = Wait(parsePen, Silver, 1, SolidLine, RoundCap)
  minorGridPen = Wait(parsePen, Silver, 1, Qt.PenStyle.DashLine, RoundCap)
  borderPen = Wait(parsePen, Black, 1, SolidLine, RoundCap)
  tickLabelFont = Wait(parseFont, 'Cambria', 6, )
  tickLabelPen = Wait(parsePen, Black, 1, SolidLine, RoundCap)
  fillBrush = Wait(parseBrush, AliceBlue, SolidFill)
  tickFillBrush = Wait(parseBrush, OliveDrab, SolidFill)

  def __init__(self, *args) -> None:
    self.dataEcho = DataEcho(128)

  def callback(self, data: Any) -> None:
    """Callback for data update"""
    self.dataEcho.append(data.data)

  def _paintOp(self, event: QPaintEvent, painter: QPainter) -> None:
    """Applies the paint operation"""
    widget = painter.device()
    minorGridPen = parsePen(Silver, 1, Qt.PenStyle.DashLine)
    majorGridPen = parsePen(Silver, 1, Qt.PenStyle.SolidLine)
    borderPen = parsePen(Black, 1, Qt.PenStyle.SolidLine)
    rect = event.rect()
    width = rect.width()
    height = rect.height()
    gridSquareSize = 8
    gridWidth = width - (width % gridSquareSize)
    gridHeight = height - (height % gridSquareSize)
    leftGrid = rect.left() + ceil((width % gridSquareSize) / 2)
    rightGrid = rect.right() - floor((width % gridSquareSize) / 2)
    topGrid = rect.top() + ceil((height % gridSquareSize) / 2)
    bottomGrid = rect.bottom() - floor((height % gridSquareSize) / 2)
    painter.setBrush(emptyBrush())
    painter.setPen(borderPen)
    painter.drawRect(rect)

    data = self.dataEcho @ event.rect()
    painter.setPen(self.pointPen)
    for (i, point) in enumerate(data):
      painter.drawPoint(point)

    painter.setPen(minorGridPen)
    t = rect.left() + ceil((width % gridSquareSize) / 2)
    x = rect.top() + ceil((height % gridSquareSize) / 2)
    _c = 0
    maxRight = rect.right() - floor((width % gridSquareSize) / 2)
    maxBottom = rect.bottom() - floor((height % gridSquareSize) / 2)
    while t < maxRight or x < maxBottom:
      if not _c % 5:
        painter.setPen(majorGridPen)
      if t < rect.right() - floor((width % gridSquareSize) / 2):
        painter.drawLine(t, rect.top(), t, rect.bottom())
      if x < rect.bottom() - floor((height % gridSquareSize) / 2):
        painter.drawLine(rect.left(), x, rect.right(), x)
      if not _c % 5:
        painter.setPen(minorGridPen)
      t += gridSquareSize
      x += gridSquareSize
      _c += 1

    bottomMargin = painter.viewport().height() - rect.bottom()
    horizontalTickHeight = bottomMargin - 4
    horizontalTickWidth = rect.width()
    horizontalTickLeft = rect.left()
    horizontalTickVLine = rect.bottom() + bottomMargin / 2
    hTicks = self.dataEcho.horizontalTicks(8, rect.width(), rect.left())
    for tick in hTicks:
      label = '%.3f' % tick[0]
      place = tick[1]

  def paintOpFill(self, event: QPaintEvent, painter: QPainter) -> None:
    """Fills background"""
    rect = event.rect()
    painter.setPen(emptyPen())
    painter.setBrush(self.fillBrush)
    painter.drawRect(rect)

  def paintOpBorder(self, event: QPaintEvent, painter: QPainter) -> None:
    """Draws border on background"""
    rect = event.rect()
    painter.setPen(self.borderPen)
    painter.setBrush(emptyBrush())
    painter.drawRect(rect)

  def paintOpGrid(self, event: QPaintEvent, painter: QPainter) -> None:
    """Paints the gridlines"""
    rect = event.rect()
    width = rect.width()
    height = rect.height()
    gridSquareSize = 8
    gridWidth = width - (width % gridSquareSize)
    gridHeight = height - (height % gridSquareSize)
    leftGrid = rect.left() + ceil((width % gridSquareSize) / 2)
    rightGrid = rect.right() - floor((width % gridSquareSize) / 2)
    topGrid = rect.top() + ceil((height % gridSquareSize) / 2)
    bottomGrid = rect.bottom() - floor((height % gridSquareSize) / 2)
    painter.setPen(self.minorGridPen)
    t = rect.left() + ceil((width % gridSquareSize) / 2)
    x = rect.top() + ceil((height % gridSquareSize) / 2)
    _c = 0
    maxRight = rect.right() - floor((width % gridSquareSize) / 2)
    maxBottom = rect.bottom() - floor((height % gridSquareSize) / 2)
    while t < maxRight or x < maxBottom:
      if not _c % 5:
        painter.setPen(self.majorGridPen)
      if t < rect.right() - floor((width % gridSquareSize) / 2):
        painter.drawLine(t, rect.top(), t, rect.bottom())
      if x < rect.bottom() - floor((height % gridSquareSize) / 2):
        painter.drawLine(rect.left(), x, rect.right(), x)
      if not _c % 5:
        painter.setPen(self.minorGridPen)
      t += gridSquareSize
      x += gridSquareSize
      _c += 1

  def paintOpData(self, event: QPaintEvent, painter: QPainter) -> None:
    """Paints the data points"""
    rect = event.rect()
    data = self.dataEcho @ event.rect()
    painter.setPen(self.pointPen)
    for (i, point) in enumerate(data):
      painter.drawPoint(point)

  def paintOpHTick(self, event: QPaintEvent, painter: QPainter) -> None:
    """Paints the ticks"""
    painter.setFont(self.tickLabelFont)
    rect = event.rect()
    bottomMargin = painter.viewport().height() - rect.bottom()
    horizontalTickHeight = bottomMargin - 4
    horizontalTickWidth = rect.width()
    horizontalTickLeft = rect.left()
    horizontalTickVLine = rect.bottom() + bottomMargin / 2
    hTicks = self.dataEcho.horizontalTicks(8, rect.width(), rect.left())
    labelSize = painter.fontMetrics().boundingRect('%.2fs' % 69).size()
    for tick in hTicks:
      label = '%.2fs' % tick[0]
      labelRect = QRect(QPoint(0, 0), labelSize)
      labelRect.moveCenter(QPoint(tick[1], horizontalTickVLine))
      painter.setPen(self.tickLabelPen)
      painter.setFont(self.tickLabelFont)
      tickTop = labelRect.top()
      rect.bottom()
      tickLine = QLine(tick[1], rect.bottom(), tick[1], tickTop)
      painter.drawLine(tickLine)
      painter.drawText(labelRect, label)

  def paintOpVTick(self, event: QPaintEvent, painter: QPainter) -> None:
    """Paints the ticks"""
    painter.setFont(self.tickLabelFont)
    labelSize = painter.fontMetrics().boundingRect('%.2f' % 420).size()
    rect = event.rect()
    verticalTickLine = rect.left() - labelSize.width()
    vTicks = self.dataEcho.verticalTicks(4, rect.height(), rect.top())
    for tick in vTicks:
      label = '%.2f' % tick[0]
      labelRect = QRect(QPoint(0, 0), labelSize)
      labelRect.moveCenter(QPoint(verticalTickLine, tick[1]))
      painter.setPen(self.tickLabelPen)
      painter.setFont(self.tickLabelFont)
      tickLine = QLine(rect.left(), tick[1], labelRect.right(), tick[1])
      painter.setBrush(self.tickFillBrush)
      painter.drawLine(tickLine)
      painter.drawText(labelRect, label)

  def paintOp(self, event: QPaintEvent, painter: QPainter) -> None:
    """Applies the paint operation"""
    self.paintOpFill(event, painter)
    self.paintOpBorder(event, painter)
    self.paintOpGrid(event, painter)
    self.paintOpData(event, painter)
    self.paintOpHTick(event, painter)
    self.paintOpVTick(event, painter)



"""DialogField provides a descriptor class for dialog boxes. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
from typing import Any, Callable

from PySide6.QtCore import Signal, QObject
from PySide6.QtWidgets import QDialog
from vistutils.waitaminute import typeMsg


class AbstractDialogField(QObject):
  """DialogField provides a descriptor class for dialog boxes. """

  __field_name__: str = None
  __field_owner__: type = None
  __positional_args__: list = None
  __keyword_args__: dict = None

  finished = Signal(int)

  def __set_name__(self, owner: type, name: str) -> None:
    """Invoked automatically after the owner is created, but before the
    owner is initiated. Let Owner be a class derived from a custom
    metaclass Meta, let Field be a descriptor derived from type and let
    field be an instance of Field owned by owner. Then the following
    ordering of events occur:
    1.  namespace = Meta.__prepare__(mcls, name, bases, ) creates the
        namespace object
    2.  field = Field() creates the field instance
    3   Owner = type.__new__(mcls, name, bases, namespace) creates the
        owner class
    4.  field.__set_name__(Owner, name) sets the field name and owner
    5.  type.__init__(Owner, name, bases, namespace) initiates the owner
        class"""
    self.__field_name__ = name
    self.__field_owner__ = owner
    setattr(owner, self._getSlotName(), Signal())

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the descriptor."""
    QObject.__init__(self, *args, **kwargs)
    self.__positional_args__ = [*args, ]
    self.__keyword_args__ = {**kwargs, }

  def getArgs(self) -> list:
    """Returns the positional arguments."""
    return self.__positional_args__

  def getKwargs(self) -> dict:
    """Returns the keyword arguments."""
    return self.__keyword_args__

  def _getFieldName(self) -> str:
    """Getter-function for getting the field name."""
    if self.__field_name__ is None:
      e = """Field name not defined!"""
      raise AttributeError(e)
    if isinstance(self.__field_name__, str):
      return self.__field_name__
    e = typeMsg('__field_name__', self.__field_name__, str)
    raise TypeError(e)

  def _getFieldOwner(self) -> type:
    """Getter-function for getting the field owner."""
    if self.__field_owner__ is None:
      e = """Field owner not defined!"""
      raise AttributeError(e)
    if isinstance(self.__field_owner__, type):
      return self.__field_owner__
    e = typeMsg('__field_owner__', self.__field_owner__, type)
    raise TypeError(e)

  def _getPrivateName(self) -> str:
    """Getter-function for getting the private name."""
    return '_%s' % self._getFieldName()

  def _getSlotName(self) -> str:
    """Getter-function for getting the slot name."""
    return '__%s_finished__' % self._getFieldName()

  def __get__(self, instance: object, owner: type) -> Any:
    """Returns the class type."""
    if instance is None:
      return self
    pvtName = self._getPrivateName()
    args = [instance, *self.getArgs()]
    kwargs = self.getKwargs()
    dialog = self._createDialog(*args, **kwargs)
    dialog.finished.connect()
    return getattr(instance, pvtName, None)

  @abstractmethod
  def _createDialog(self, instance: Any, callMeMaybe: Callable) -> QDialog:
    """Creates the dialog box."""



"""The RightNowField provides a StaticField return the epoch the moment
the __getter__ is invoked. If accessed on the owner class, the descriptor
returns itself, if accessed on any instance, it returns the time. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import time
from typing import Any

from vistutils.fields import StaticField
from vistutils.waitaminute import typeMsg


class RightNowField(StaticField):
  """The RightNowField provides a StaticField return the epoch the moment
    the __getter__ is invoked. If accessed on the owner class, the descriptor
    returns itself, if accessed on any instance, it returns the time. """

  def __init__(self, *args, **kwargs) -> None:
    StaticField.__init__(self, None)

  def __get__(self, instance: object, owner: type) -> Any:
    """Getter-function implementation."""
    if instance is None:
      return self
    if getattr(owner, '__zero_time__', None) is None:
      e = """The owner class does not have a __zero_time__ attribute!"""
      raise AttributeError(e)
    zeroTime = getattr(owner, '__zero_time__')
    if isinstance(zeroTime, float):
      return time.time() - zeroTime
    e = typeMsg('zeroTime', zeroTime, float)
    raise TypeError(e)

  def __set_name__(self, owner: type, name: str) -> None:
    """Invoked automatically when the owner class is created."""
    StaticField.__set_name__(self, owner, name)
    setattr(owner, '__zero_time__', time.time())



"""TextRect places text on the given rectangle"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QPaintEvent, QPainter, QFontMetricsF, QPaintDevice
from vistutils.waitaminute import typeMsg

from ezros.gui.factories import textPen, parseFont, parseColor
from ezros.gui.paint import AbstractPaint
from ezros.gui.shortnames import Black
from ezros.rosutils import Wait


class TextRect(AbstractPaint):
  """TextRect places text on the given rectangle"""

  __self_font_metrics__ = None
  __recent_painter__ = None

  textLine = Wait(textPen, )
  textFont = Wait(parseFont, 'Arial', 10)
  fontMetrics = Wait()

  def paintOp(self,
              event: QPaintEvent,
              painter: QPainter,
              text: str = None) -> None:
    """Prints text found in the painter device"""
    if text is None:
      if getattr(painter.device(), 'innerText', None) is not None:
        text = painter.device().innerText
        if not isinstance(text, str):
          text = str(text)
    painter.setPen(self.textLine)
    painter.setFont(self.textFont)
    painter.drawText(event.rect(), text)

  @fontMetrics.CREATE
  def _getFontMetrics(self, **kwargs) -> QFontMetricsF:
    """Getter for the font metrics"""
    if self.__self_font_metrics__ is None:
      if kwargs.get('recursion', False):
        raise RecursionError
      self.__self_font_metrics__ = QFontMetricsF(self.textFont)
      return self._getFontMetrics(recursion=True)
    return self.__self_font_metrics__

  def __init__(self, *args, **kwargs) -> None:
    color, font = None, None
    color = parseColor(*args, **kwargs, strict=False)
    font = parseFont(*args, **kwargs, strict=False)



"""This file contains named colors for use in the GUI."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QColor


class ColorSpace:
  """Color space class"""
  AliceBlue = QColor(240, 248, 255)
  AntiqueWhite = QColor(250, 235, 215)
  Aqua = QColor(0, 255, 255)
  Aquamarine = QColor(127, 255, 212)
  Azure = QColor(240, 255, 255)
  Beige = QColor(245, 245, 220)
  Bisque = QColor(255, 228, 196)
  Black = QColor(0, 0, 0)
  BlanchedAlmond = QColor(255, 235, 205)
  Blue = QColor(0, 0, 255)
  BlueViolet = QColor(138, 43, 226)
  Brown = QColor(165, 42, 42)
  BurlyWood = QColor(222, 184, 135)
  CadetBlue = QColor(95, 158, 160)
  Chartreuse = QColor(127, 255, 0)
  Chocolate = QColor(210, 105, 30)
  Coral = QColor(255, 127, 80)
  CornflowerBlue = QColor(100, 149, 237)
  Cornsilk = QColor(255, 248, 220)
  Crimson = QColor(220, 20, 60)
  Cyan = QColor(0, 255, 255)
  DarkBlue = QColor(0, 0, 139)
  DarkCyan = QColor(0, 139, 139)
  DarkGoldenRod = QColor(184, 134, 11)
  DarkGray = QColor(169, 169, 169)
  DarkGreen = QColor(0, 100, 0)
  DarkKhaki = QColor(189, 183, 107)
  DarkMagenta = QColor(139, 0, 139)
  DarkOliveGreen = QColor(85, 107, 47)
  DarkOrange = QColor(255, 140, 0)
  DarkOrchid = QColor(153, 50, 204)
  DarkRed = QColor(139, 0, 0)
  DarkSalmon = QColor(233, 150, 122)
  DarkSeaGreen = QColor(143, 188, 143)
  DarkSlateBlue = QColor(72, 61, 139)
  DarkSlateGray = QColor(47, 79, 79)
  DarkTurquoise = QColor(0, 206, 209)
  DarkViolet = QColor(148, 0, 211)
  DeepPink = QColor(255, 20, 147)
  DeepSkyBlue = QColor(0, 191, 255)
  DimGray = QColor(105, 105, 105)
  DodgerBlue = QColor(30, 144, 255)
  FireBrick = QColor(178, 34, 34)
  FloralWhite = QColor(255, 250, 240)
  ForestGreen = QColor(34, 139, 34)
  Fuchsia = QColor(255, 0, 255)
  Gainsboro = QColor(220, 220, 220)
  GhostWhite = QColor(248, 248, 255)
  Gold = QColor(255, 215, 0)
  GoldenRod = QColor(218, 165, 32)
  Gray = QColor(128, 128, 128)
  Green = QColor(0, 128, 0)
  GreenYellow = QColor(173, 255, 47)
  HoneyDew = QColor(240, 255, 240)
  HotPink = QColor(255, 105, 180)
  IndianRed = QColor(205, 92, 92)
  Indigo = QColor(75, 0, 130)
  Ivory = QColor(255, 255, 240)
  Khaki = QColor(240, 230, 140)
  Lavender = QColor(230, 230, 250)
  LavenderBlush = QColor(255, 240, 245)
  LawnGreen = QColor(124, 252, 0)
  LemonChiffon = QColor(255, 250, 205)
  LightBlue = QColor(173, 216, 230)
  LightCoral = QColor(240, 128, 128)
  LightCyan = QColor(224, 255, 255)
  LightGoldenRodYellow = QColor(250, 250, 210)
  LightGray = QColor(211, 211, 211)
  LightGreen = QColor(144, 238, 144)
  LightPink = QColor(255, 182, 193)
  LightSalmon = QColor(255, 160, 122)
  LightSeaGreen = QColor(32, 178, 170)
  LightSkyBlue = QColor(135, 206, 250)
  LightSlateGray = QColor(119, 136, 153)
  LightSteelBlue = QColor(176, 224, 230)
  LightYellow = QColor(255, 255, 224)
  Lime = QColor(0, 255, 0)
  LimeGreen = QColor(50, 205, 50)
  Linen = QColor(250, 240, 230)
  Magenta = QColor(255, 0, 255)
  Maroon = QColor(128, 0, 0)
  MediumAquaMarine = QColor(102, 205, 170)
  MediumBlue = QColor(0, 0, 205)
  MediumOrchid = QColor(186, 85, 211)
  MediumPurple = QColor(147, 112, 219)
  MediumSeaGreen = QColor(60, 179, 113)
  MediumSlateBlue = QColor(123, 104, 238)
  MediumSpringGreen = QColor(0, 250, 154)
  MediumTurquoise = QColor(72, 209, 204)
  MediumVioletRed = QColor(199, 21, 133)
  MidnightBlue = QColor(25, 25, 112)
  MintCream = QColor(245, 255, 250)
  MistyRose = QColor(255, 228, 225)
  Moccasin = QColor(255, 228, 181)
  NavajoWhite = QColor(255, 222, 173)
  Navy = QColor(0, 0, 128)
  OldLace = QColor(253, 245, 230)
  Olive = QColor(128, 128, 0)
  OliveDrab = QColor(107, 142, 35)
  Orange = QColor(255, 165, 0)
  OrangeRed = QColor(255, 69, 0)
  Orchid = QColor(218, 112, 214)
  PaleGoldenRod = QColor(238, 232, 170)
  PaleGreen = QColor(152, 251, 152)
  PaleTurquoise = QColor(175, 238, 238)
  PaleVioletRed = QColor(219, 112, 147)
  PapayaWhip = QColor(255, 239, 213)
  PeachPuff = QColor(255, 218, 185)
  Peru = QColor(205, 133, 63)
  Pink = QColor(255, 192, 203)
  Plum = QColor(221, 160, 221)
  PowderBlue = QColor(176, 224, 230)
  Purple = QColor(128, 0, 128)
  RebeccaPurple = QColor(102, 51, 153)
  Red = QColor(255, 0, 0)
  RosyBrown = QColor(188, 143, 143)
  RoyalBlue = QColor(65, 105, 225)
  SaddleBrown = QColor(139, 69, 19)
  Salmon = QColor(250, 128, 114)
  SandyBrown = QColor(244, 164, 96)
  SeaGreen = QColor(46, 139, 87)
  SeaShell = QColor(255, 245, 238)
  Sienna = QColor(160, 82, 45)
  Silver = QColor(192, 192, 192)
  SkyBlue = QColor(135, 206, 235)
  SlateBlue = QColor(106, 90, 205)
  SlateGray = QColor(112, 128, 144)
  Snow = QColor(255, 250, 250)
  SpringGreen = QColor(0, 255, 127)
  SteelBlue = QColor(70, 130, 180)
  Tan = QColor(210, 180, 140)
  Teal = QColor(0, 128, 128)
  Thistle = QColor(216, 191, 216)
  Tomato = QColor(255, 99, 71)
  Turquoise = QColor(64, 224, 208)
  Violet = QColor(238, 130, 238)
  Wheat = QColor(245, 222, 179)
  White = QColor(255, 255, 255)
  WhiteSmoke = QColor(245, 245, 245)
  Yellow = QColor(255, 255, 0)
  YellowGreen = QColor(154, 205, 50)


AliceBlue = QColor(240, 248, 255)
AntiqueWhite = QColor(250, 235, 215)
Aqua = QColor(0, 255, 255)
Aquamarine = QColor(127, 255, 212)
Azure = QColor(240, 255, 255)
Beige = QColor(245, 245, 220)
Bisque = QColor(255, 228, 196)
Black = QColor(0, 0, 0)
BlanchedAlmond = QColor(255, 235, 205)
Blue = QColor(0, 0, 255)
BlueViolet = QColor(138, 43, 226)
Brown = QColor(165, 42, 42)
BurlyWood = QColor(222, 184, 135)
CadetBlue = QColor(95, 158, 160)
Chartreuse = QColor(127, 255, 0)
Chocolate = QColor(210, 105, 30)
Coral = QColor(255, 127, 80)
CornflowerBlue = QColor(100, 149, 237)
Cornsilk = QColor(255, 248, 220)
Crimson = QColor(220, 20, 60)
Cyan = QColor(0, 255, 255)
DarkBlue = QColor(0, 0, 139)
DarkCyan = QColor(0, 139, 139)
DarkGoldenRod = QColor(184, 134, 11)
DarkGray = QColor(169, 169, 169)
DarkGreen = QColor(0, 100, 0)
DarkKhaki = QColor(189, 183, 107)
DarkMagenta = QColor(139, 0, 139)
DarkOliveGreen = QColor(85, 107, 47)
DarkOrange = QColor(255, 140, 0)
DarkOrchid = QColor(153, 50, 204)
DarkRed = QColor(139, 0, 0)
DarkSalmon = QColor(233, 150, 122)
DarkSeaGreen = QColor(143, 188, 143)
DarkSlateBlue = QColor(72, 61, 139)
DarkSlateGray = QColor(47, 79, 79)
DarkTurquoise = QColor(0, 206, 209)
DarkViolet = QColor(148, 0, 211)
DeepPink = QColor(255, 20, 147)
DeepSkyBlue = QColor(0, 191, 255)
DimGray = QColor(105, 105, 105)
DodgerBlue = QColor(30, 144, 255)
FireBrick = QColor(178, 34, 34)
FloralWhite = QColor(255, 250, 240)
ForestGreen = QColor(34, 139, 34)
Fuchsia = QColor(255, 0, 255)
Gainsboro = QColor(220, 220, 220)
GhostWhite = QColor(248, 248, 255)
Gold = QColor(255, 215, 0)
GoldenRod = QColor(218, 165, 32)
Gray = QColor(128, 128, 128)
Green = QColor(0, 128, 0)
GreenYellow = QColor(173, 255, 47)
HoneyDew = QColor(240, 255, 240)
HotPink = QColor(255, 105, 180)
IndianRed = QColor(205, 92, 92)
Indigo = QColor(75, 0, 130)
Ivory = QColor(255, 255, 240)
Khaki = QColor(240, 230, 140)
Lavender = QColor(230, 230, 250)
LavenderBlush = QColor(255, 240, 245)
LawnGreen = QColor(124, 252, 0)
LemonChiffon = QColor(255, 250, 205)
LightBlue = QColor(173, 216, 230)
LightCoral = QColor(240, 128, 128)
LightCyan = QColor(224, 255, 255)
LightGoldenRodYellow = QColor(250, 250, 210)
LightGray = QColor(211, 211, 211)
LightGreen = QColor(144, 238, 144)
LightPink = QColor(255, 182, 193)
LightSalmon = QColor(255, 160, 122)
LightSeaGreen = QColor(32, 178, 170)
LightSkyBlue = QColor(135, 206, 250)
LightSlateGray = QColor(119, 136, 153)
LightSteelBlue = QColor(176, 224, 230)
LightYellow = QColor(255, 255, 224)
Lime = QColor(0, 255, 0)
LimeGreen = QColor(50, 205, 50)
Linen = QColor(250, 240, 230)
Magenta = QColor(255, 0, 255)
Maroon = QColor(128, 0, 0)
MediumAquaMarine = QColor(102, 205, 170)
MediumBlue = QColor(0, 0, 205)
MediumOrchid = QColor(186, 85, 211)
MediumPurple = QColor(147, 112, 219)
MediumSeaGreen = QColor(60, 179, 113)
MediumSlateBlue = QColor(123, 104, 238)
MediumSpringGreen = QColor(0, 250, 154)
MediumTurquoise = QColor(72, 209, 204)
MediumVioletRed = QColor(199, 21, 133)
MidnightBlue = QColor(25, 25, 112)
MintCream = QColor(245, 255, 250)
MistyRose = QColor(255, 228, 225)
Moccasin = QColor(255, 228, 181)
NavajoWhite = QColor(255, 222, 173)
Navy = QColor(0, 0, 128)
OldLace = QColor(253, 245, 230)
Olive = QColor(128, 128, 0)
OliveDrab = QColor(107, 142, 35)
Orange = QColor(255, 165, 0)
OrangeRed = QColor(255, 69, 0)
Orchid = QColor(218, 112, 214)
PaleGoldenRod = QColor(238, 232, 170)
PaleGreen = QColor(152, 251, 152)
PaleTurquoise = QColor(175, 238, 238)
PaleVioletRed = QColor(219, 112, 147)
PapayaWhip = QColor(255, 239, 213)
PeachPuff = QColor(255, 218, 185)
Peru = QColor(205, 133, 63)
Pink = QColor(255, 192, 203)
Plum = QColor(221, 160, 221)
PowderBlue = QColor(176, 224, 230)
Purple = QColor(128, 0, 128)
RebeccaPurple = QColor(102, 51, 153)
Red = QColor(255, 0, 0)
RosyBrown = QColor(188, 143, 143)
RoyalBlue = QColor(65, 105, 225)
SaddleBrown = QColor(139, 69, 19)
Salmon = QColor(250, 128, 114)
SandyBrown = QColor(244, 164, 96)
SeaGreen = QColor(46, 139, 87)
SeaShell = QColor(255, 245, 238)
Sienna = QColor(160, 82, 45)
Silver = QColor(192, 192, 192)
SkyBlue = QColor(135, 206, 235)
SlateBlue = QColor(106, 90, 205)
SlateGray = QColor(112, 128, 144)
Snow = QColor(255, 250, 250)
SpringGreen = QColor(0, 255, 127)
SteelBlue = QColor(70, 130, 180)
Tan = QColor(210, 180, 140)
Teal = QColor(0, 128, 128)
Thistle = QColor(216, 191, 216)
Tomato = QColor(255, 99, 71)
Turquoise = QColor(64, 224, 208)
Violet = QColor(238, 130, 238)
Wheat = QColor(245, 222, 179)
White = QColor(255, 255, 255)
WhiteSmoke = QColor(245, 245, 245)
Yellow = QColor(255, 255, 0)
YellowGreen = QColor(154, 205, 50)

greyDict = {
  'Black'        : QColor(0, 0, 0, ),
  'DimGray'      : QColor(105, 105, 105, ),
  'Gray'         : QColor(128, 128, 128, ),
  'DarkGray'     : QColor(169, 169, 169, ),
  'Silver'       : QColor(192, 192, 192, ),
  'LightGray'    : QColor(211, 211, 211, ),
  'Gainsboro'    : QColor(220, 220, 220, ),
  'HoneyDew'     : QColor(240, 255, 240, ),
  'WhiteSmoke'   : QColor(245, 245, 245, ),
  'LavenderBlush': QColor(255, 240, 245, ),
  'AliceBlue'    : QColor(240, 248, 255, ),
  'FloralWhite'  : QColor(255, 250, 240, ),
  'Azure'        : QColor(240, 255, 255, ),
  'Ivory'        : QColor(255, 255, 240, ),
  'MintCream'    : QColor(245, 255, 250, ),
  'GhostWhite'   : QColor(248, 248, 255, ),
  'Snow'         : QColor(255, 250, 250, ),
  'White'        : QColor(255, 255, 255, ),
}

colorDict = {
  'AliceBlue'           : QColor(240, 248, 255, ),
  'AntiqueWhite'        : QColor(250, 235, 215, ),
  'Aqua'                : QColor(0, 255, 255, ),
  'Aquamarine'          : QColor(127, 255, 212, ),
  'Azure'               : QColor(240, 255, 255, ),
  'Beige'               : QColor(245, 245, 220, ),
  'Bisque'              : QColor(255, 228, 196, ),
  'Black'               : QColor(0, 0, 0, ),
  'BlanchedAlmond'      : QColor(255, 235, 205, ),
  'Blue'                : QColor(0, 0, 255, ),
  'BlueViolet'          : QColor(138, 43, 226, ),
  'Brown'               : QColor(165, 42, 42, ),
  'BurlyWood'           : QColor(222, 184, 135, ),
  'CadetBlue'           : QColor(95, 158, 160, ),
  'Chartreuse'          : QColor(127, 255, 0, ),
  'Chocolate'           : QColor(210, 105, 30, ),
  'Coral'               : QColor(255, 127, 80, ),
  'CornflowerBlue'      : QColor(100, 149, 237, ),
  'Cornsilk'            : QColor(255, 248, 220, ),
  'Crimson'             : QColor(220, 20, 60, ),
  'Cyan'                : QColor(0, 255, 255, ),
  'DarkBlue'            : QColor(0, 0, 139, ),
  'DarkCyan'            : QColor(0, 139, 139, ),
  'DarkGoldenRod'       : QColor(184, 134, 11, ),
  'DarkGray'            : QColor(169, 169, 169, ),
  'DarkGreen'           : QColor(0, 100, 0, ),
  'DarkKhaki'           : QColor(189, 183, 107, ),
  'DarkMagenta'         : QColor(139, 0, 139, ),
  'DarkOliveGreen'      : QColor(85, 107, 47, ),
  'DarkOrange'          : QColor(255, 140, 0, ),
  'DarkOrchid'          : QColor(153, 50, 204, ),
  'DarkRed'             : QColor(139, 0, 0, ),
  'DarkSalmon'          : QColor(233, 150, 122, ),
  'DarkSeaGreen'        : QColor(143, 188, 143, ),
  'DarkSlateBlue'       : QColor(72, 61, 139, ),
  'DarkSlateGray'       : QColor(47, 79, 79, ),
  'DarkTurquoise'       : QColor(0, 206, 209, ),
  'DarkViolet'          : QColor(148, 0, 211, ),
  'DeepPink'            : QColor(255, 20, 147, ),
  'DeepSkyBlue'         : QColor(0, 191, 255, ),
  'DimGray'             : QColor(105, 105, 105, ),
  'DodgerBlue'          : QColor(30, 144, 255, ),
  'FireBrick'           : QColor(178, 34, 34, ),
  'FloralWhite'         : QColor(255, 250, 240, ),
  'ForestGreen'         : QColor(34, 139, 34, ),
  'Fuchsia'             : QColor(255, 0, 255, ),
  'Gainsboro'           : QColor(220, 220, 220, ),
  'GhostWhite'          : QColor(248, 248, 255, ),
  'Gold'                : QColor(255, 215, 0, ),
  'GoldenRod'           : QColor(218, 165, 32, ),
  'Gray'                : QColor(128, 128, 128, ),
  'Green'               : QColor(0, 128, 0, ),
  'GreenYellow'         : QColor(173, 255, 47, ),
  'HoneyDew'            : QColor(240, 255, 240, ),
  'HotPink'             : QColor(255, 105, 180, ),
  'IndianRed'           : QColor(205, 92, 92, ),
  'Indigo'              : QColor(75, 0, 130, ),
  'Ivory'               : QColor(255, 255, 240, ),
  'Khaki'               : QColor(240, 230, 140, ),
  'Lavender'            : QColor(230, 230, 250, ),
  'LavenderBlush'       : QColor(255, 240, 245, ),
  'LawnGreen'           : QColor(124, 252, 0, ),
  'LemonChiffon'        : QColor(255, 250, 205, ),
  'LightBlue'           : QColor(173, 216, 230, ),
  'LightCoral'          : QColor(240, 128, 128, ),
  'LightCyan'           : QColor(224, 255, 255, ),
  'LightGoldenRodYellow': QColor(250, 250, 210, ),
  'LightGray'           : QColor(211, 211, 211, ),
  'LightGreen'          : QColor(144, 238, 144, ),
  'LightPink'           : QColor(255, 182, 193, ),
  'LightSalmon'         : QColor(255, 160, 122, ),
  'LightSeaGreen'       : QColor(32, 178, 170, ),
  'LightSkyBlue'        : QColor(135, 206, 250, ),
  'LightSlateGray'      : QColor(119, 136, 153, ),
  'LightSteelBlue'      : QColor(176, 224, 230, ),
  'LightYellow'         : QColor(255, 255, 224, ),
  'Lime'                : QColor(0, 255, 0, ),
  'LimeGreen'           : QColor(50, 205, 50, ),
  'Linen'               : QColor(250, 240, 230, ),
  'Magenta'             : QColor(255, 0, 255, ),
  'Maroon'              : QColor(128, 0, 0, ),
  'MediumAquaMarine'    : QColor(102, 205, 170, ),
  'MediumBlue'          : QColor(0, 0, 205, ),
  'MediumOrchid'        : QColor(186, 85, 211, ),
  'MediumPurple'        : QColor(147, 112, 219, ),
  'MediumSeaGreen'      : QColor(60, 179, 113, ),
  'MediumSlateBlue'     : QColor(123, 104, 238, ),
  'MediumSpringGreen'   : QColor(0, 250, 154, ),
  'MediumTurquoise'     : QColor(72, 209, 204, ),
  'MediumVioletRed'     : QColor(199, 21, 133, ),
  'MidnightBlue'        : QColor(25, 25, 112, ),
  'MintCream'           : QColor(245, 255, 250, ),
  'MistyRose'           : QColor(255, 228, 225, ),
  'Moccasin'            : QColor(255, 228, 181, ),
  'NavajoWhite'         : QColor(255, 222, 173, ),
  'Navy'                : QColor(0, 0, 128, ),
  'OldLace'             : QColor(253, 245, 230, ),
  'Olive'               : QColor(128, 128, 0, ),
  'OliveDrab'           : QColor(107, 142, 35, ),
  'Orange'              : QColor(255, 165, 0, ),
  'OrangeRed'           : QColor(255, 69, 0, ),
  'Orchid'              : QColor(218, 112, 214, ),
  'PaleGoldenRod'       : QColor(238, 232, 170, ),
  'PaleGreen'           : QColor(152, 251, 152, ),
  'PaleTurquoise'       : QColor(175, 238, 238, ),
  'PaleVioletRed'       : QColor(219, 112, 147, ),
  'PapayaWhip'          : QColor(255, 239, 213, ),
  'PeachPuff'           : QColor(255, 218, 185, ),
  'Peru'                : QColor(205, 133, 63, ),
  'Pink'                : QColor(255, 192, 203, ),
  'Plum'                : QColor(221, 160, 221, ),
  'PowderBlue'          : QColor(176, 224, 230, ),
  'Purple'              : QColor(128, 0, 128, ),
  'RebeccaPurple'       : QColor(102, 51, 153, ),
  'Red'                 : QColor(255, 0, 0, ),
  'RosyBrown'           : QColor(188, 143, 143, ),
  'RoyalBlue'           : QColor(65, 105, 225, ),
  'SaddleBrown'         : QColor(139, 69, 19, ),
  'Salmon'              : QColor(250, 128, 114, ),
  'SandyBrown'          : QColor(244, 164, 96, ),
  'SeaGreen'            : QColor(46, 139, 87, ),
  'SeaShell'            : QColor(255, 245, 238, ),
  'Sienna'              : QColor(160, 82, 45, ),
  'Silver'              : QColor(192, 192, 192, ),
  'SkyBlue'             : QColor(135, 206, 235, ),
  'SlateBlue'           : QColor(106, 90, 205, ),
  'SlateGray'           : QColor(112, 128, 144, ),
  'Snow'                : QColor(255, 250, 250, ),
  'SpringGreen'         : QColor(0, 255, 127, ),
  'SteelBlue'           : QColor(70, 130, 180, ),
  'Tan'                 : QColor(210, 180, 140, ),
  'Teal'                : QColor(0, 128, 128, ),
  'Thistle'             : QColor(216, 191, 216, ),
  'Tomato'              : QColor(255, 99, 71, ),
  'Turquoise'           : QColor(64, 224, 208, ),
  'Violet'              : QColor(238, 130, 238, ),
  'Wheat'               : QColor(245, 222, 179, ),
  'White'               : QColor(255, 255, 255, ),
  'WhiteSmoke'          : QColor(245, 245, 245, ),
  'Yellow'              : QColor(255, 255, 0, ),
  'YellowGreen'         : QColor(154, 205, 50, ),
}

__all__ = [
  'ColorSpace',
  'colorDict',
  'AliceBlue',
  'AntiqueWhite',
  'Aqua',
  'Aquamarine',
  'Azure',
  'Beige',
  'Bisque',
  'Black',
  'BlanchedAlmond',
  'Blue',
  'BlueViolet',
  'Brown',
  'BurlyWood',
  'CadetBlue',
  'Chartreuse',
  'Chocolate',
  'Coral',
  'CornflowerBlue',
  'Cornsilk',
  'Crimson',
  'Cyan',
  'DarkBlue',
  'DarkCyan',
  'DarkGoldenRod',
  'DarkGray',
  'DarkGreen',
  'DarkKhaki',
  'DarkMagenta',
  'DarkOliveGreen',
  'DarkOrange',
  'DarkOrchid',
  'DarkRed',
  'DarkSalmon',
  'DarkSeaGreen',
  'DarkSlateBlue',
  'DarkSlateGray',
  'DarkTurquoise',
  'DarkViolet',
  'DeepPink',
  'DeepSkyBlue',
  'DimGray',
  'DodgerBlue',
  'FireBrick',
  'FloralWhite',
  'ForestGreen',
  'Fuchsia',
  'Gainsboro',
  'GhostWhite',
  'Gold',
  'GoldenRod',
  'Gray',
  'Green',
  'GreenYellow',
  'HoneyDew',
  'HotPink',
  'IndianRed',
  'Indigo',
  'Ivory',
  'Khaki',
  'Lavender',
  'LavenderBlush',
  'LawnGreen',
  'LemonChiffon',
  'LightBlue',
  'LightCoral',
  'LightCyan',
  'LightGoldenRodYellow',
  'LightGray',
  'LightGreen',
  'LightPink',
  'LightSalmon',
  'LightSeaGreen',
  'LightSkyBlue',
  'LightSlateGray',
  'LightSteelBlue',
  'LightYellow',
  'Lime',
  'LimeGreen',
  'Linen',
  'Magenta',
  'Maroon',
  'MediumAquaMarine',
  'MediumBlue',
  'MediumOrchid',
  'MediumPurple',
  'MediumSeaGreen',
  'MediumSlateBlue',
  'MediumSpringGreen',
  'MediumTurquoise',
  'MediumVioletRed',
  'MidnightBlue',
  'MintCream',
  'MistyRose',
  'Moccasin',
  'NavajoWhite',
  'Navy',
  'OldLace',
  'Olive',
  'OliveDrab',
  'Orange',
  'OrangeRed',
  'Orchid',
  'PaleGoldenRod',
  'PaleGreen',
  'PaleTurquoise',
  'PaleVioletRed',
  'PapayaWhip',
  'PeachPuff',
  'Peru',
  'Pink',
  'Plum',
  'PowderBlue',
  'Purple',
  'RebeccaPurple',
  'Red',
  'RosyBrown',
  'RoyalBlue',
  'SaddleBrown',
  'Salmon',
  'SandyBrown',
  'SeaGreen',
  'SeaShell',
  'Sienna',
  'Silver',
  'SkyBlue',
  'SlateBlue',
  'SlateGray',
  'Snow',
  'SpringGreen',
  'SteelBlue',
  'Tan',
  'Teal',
  'Thistle',
  'Tomato',
  'Turquoise',
  'Violet',
  'Wheat',
  'White',
  'WhiteSmoke',
  'Yellow',
  'YellowGreen',
]



"""PenField provides a descriptor class for instances of QPen."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from PySide6.QtGui import QPen
from vistutils.fields import unParseArgs, Wait

from vistside.core import parsePen


class PenWrap(QPen):
  """Subclass of QPen with a custom constructor enabling interaction
  with descriptors."""

  @classmethod
  def getDefault(cls, *args, **kwargs) -> PenWrap:
    """Returns the default value for the field."""
    return cls().apply((args, kwargs))

  def apply(self, value: Any) -> PenWrap:
    """Applies the value to the field."""
    args, kwargs = unParseArgs(value)
    testPen = parsePen(*args, **kwargs)
    self.setStyle(testPen.style())
    self.setColor(testPen.color())
    self.setWidth(testPen.width())
    self.setCapStyle(testPen.capStyle())
    self.setJoinStyle(testPen.joinStyle())

    return self


class PenField(Wait):
  """The PenField class provides a descriptor for instances of QPen."""

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the PenField."""
    Wait.__init__(self, PenWrap, *args, **kwargs)



"""DebugMenu provides a debug menu for development purposes."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from vistside.windows.actions import AbstractMenu


class DebugMenu(AbstractMenu):
  """DebugMenu provides a debug menu for development purposes."""

  def setupActions(self, ) -> None:
    """setupActions creates the actions for the menu."""
    self.parent().debug01Action = self.addAction('Debug01', )
    self.parent().debug02Action = self.addAction('Debug02', )
    self.parent().debug03Action = self.addAction('Debug03', )
    self.parent().debug04Action = self.addAction('Debug04', )
    self.parent().debug05Action = self.addAction('Debug05', )
    self.parent().debug06Action = self.addAction('Debug06', )
    self.parent().debug07Action = self.addAction('Debug07', )
    self.parent().debug08Action = self.addAction('Debug08', )
    self.parent().debug09Action = self.addAction('Debug09', )
    self.parent().debug10Action = self.addAction('Debug10', )
